;;; -*- Scheme -*-

;;;; Higher-level interface to the PostgreSQL RDBMS for DrScheme

;;; Author: Zbigniew Jurkiewicz <zbyszek@mimuw.edu.pl>
;;; Version: 0.1
;;;
;;;     Copyright (C) 2004-2006  Zbigniew Jurkiewicz

;;; This library is free software; you can redistribute it and/or
;;; modify it under the terms of the GNU Library General Public
;;; License as published by the Free Software Foundation; either
;;; version 2 of the License, or (at your option) any later version.
;;
;;; This library is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
;;; Library General Public License for more details.
;;
;;; You should have received a copy of the GNU Library General Public
;;; License along with this library; if not, write to the Free
;;; Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
;;;
;;; Please send suggestions and to <zbyszek@mimuw.edu.pl>.
;;;
;;; The latest version of this package should be available from
;;; 
;;;     <URL:http://duch.mimuw.edu.pl/~zbyszek/>

;;; Overview =========================================================
;;;
;;; This module lets you access the PostgreSQL object-relational DBMS
;;; from DrScheme. The code implements the public interface of Eric Marsden's
;;; socket-level library (<URL:http://purl.org/net/emarsden/home/downloads/>),
;;; but is based on our libpq library. The module is 
;;; capable of type coercions from a range of SQL types to the equivalent 
;;; Scheme type.
;;;
;;; Works with MzScheme, MrEd and DrScheme.
;;;
;;; See the README for API documentation. 

;;; Only password authentication is supported (neither crypt nor
;;; Kerberos work).

;;; This code has been tested or reported to work with
;;;
;;;   * MzScheme 209, DrScheme 209 on Linux/x86
;;;   * PostgreSQL 6.5, 7.0, 7.1.2, 7.2.

;;;; TODO ============================================================
;;;
;;; * add a mechanism for parsing user-defined types. The user should
;;;   be able to define a parse function and a type-name; we query
;;;   pg_type to get the type's OID and add the information to
;;;   *parsers*.

;;; Portability  ;***
;;
;; * MzScheme structs
;; * opt-lambda

(module mzpg mzscheme
  (provide pg-connect pg-exec pg-result pg-disconnect
           *pg-disable-type-coercion*
           pg-databases pg-tables pg-columns
           pg-backend-version
           pg-date-style set-pg-date-style
           pg-client-encoding set-pg-client-encoding
           with-pg-connection
           with-pg-transaction
           pg-for-each
           pg-notices  ;***
           pglo-create
           pglo-open
           pglo-close
           pglo-read
           pglo-write
           pglo-lseek
           pglo-tell
           pglo-unlink
           pglo-import
           pglo-export
           pg-quote)  ;***
  (require (lib "libpq-fe.scm" "libpq")
           (lib "etc.ss") (lib "date.ss") (lib "list.ss"))
  
  (define (signal-error constructor format-string)
    (lambda args  ;exception struct fields
      (raise (apply constructor
                    (apply format format-string args)
                    (current-continuation-marks)
                    args))))

  (define-struct (postgresql-error exn) ())

  #|
  (define-struct (connection-failure postgresql-error)
    (host port transport-error))
  (define signal-connection-failure
    (signal-error
     make-connection-failure
     "Couldn't connect to PostgreSQL database at ~a:~a.
Connection attempt reported ~A.
Is the postmaster running and accepting TCP connections?~%"))
  |#

  (define-struct (authentication-failure postgresql-error)
    (reason))
  (define signal-authentication-failure
    (signal-error make-authentication-failure
                  "PostgreSQL authentication failure: ~a~%"))

  (define-struct (protocol-error postgresql-error)
    (reason))
  (define signal-protocol-error
    (signal-error make-protocol-error
                  "PostgreSQL protocol error: ~a~%"))

  (define-struct (backend-error postgresql-error)
    (reason))
  (define signal-backend-error
    (signal-error make-backend-error
                  "PostgreSQL backend error: ~a~%"))

  (define +MAX_MESSAGE_LEN+ 8192)     ;libpq-fe.h  ???

  ;; An alist of (oid . parser) pairs. This is built dynamically at
  ;; initialization of the connection with the database (once generated,
  ;; the information is shared between connections).

  (define *parsers* '())

  (define *pg-client-encoding* "LATIN1")
    ;; The encoding to use for text data, for example "LATIN1", "UNICODE", 
    ;; "EUC_JP".
    ;; See <http://www.postgresql.org/docs/7.3/static/multibyte.html>.

  (define *pg-date-style* "ISO")

  (define *pg-disable-type-coercion* #f)  ;***
  ;; Non-#f disables the type coercion mechanism.
  ;; The default is #f, which means that data recovered from the
  ;; database is coerced to the corresponding Scheme type before
  ;; being returned; for example numeric data is transformed to 
  ;; Scheme numbers, and booleans to booleans.
  ;;
  ;; The coercion mechanism requires an initialization query to the
  ;; database, in order to build a table mapping type names to OIDs. This
  ;; option is provided mainly in case you wish to avoid the overhead of
  ;; this initial query. The overhead is only incurred once per session
  ;; (not per connection to the backend).

  (define-struct pgresult
    (connection
     status
     attributes
     tuples))

  (define (pg-date-style conn)
    (let ((res (pg-exec conn "SHOW datestyle")))
      (first (pg-result res 'tuple 0))))

  (define (set-pg-date-style conn new-date-style)
    (pg-exec conn "SET datestyle TO " new-date-style))

  ;; See http://www.postgresql.org/docs/7.3/static/multibyte.html
  (define (pg-client-encoding conn)
    ;; Return a string identifying the client encoding.
    (let ((res (pg-exec conn "SHOW client_encoding")))
      (first (pg-result res 'tuple 0))))

  (define (set-pg-client-encoding conn new-encoding)
    ;; Set the client_encoding.
    (pg-exec conn "SET client_encoding TO " new-encoding))

  (define-syntax with-pg-connection
    ;; Bindspec is of the form (connection open-args), where OPEN-ARGS are
    ;; as for PG-CONNECT. The database connection is bound to the variable
    ;; CONNECTION. If the connection is unsuccessful, the forms are not
    ;; evaluated. Otherwise, the BODY forms are executed, and upon
    ;; termination, normal or otherwise, the database connection is closed.
    (syntax-rules ()
      ((with-pg-connection (con open-args ...) body ...)
       (let ((con (pg-connect open-args ...)))
         (dynamic-wind void
           (lambda () body ...)
           (lambda () (when con (pg-disconnect con))))))))

  (define-syntax with-pg-transaction
    ;; Execute BODY forms in a BEGIN..END block.
    ;; If a PostgreSQL error occurs during execution of the forms, execute
    ;; a ROLLBACK command.
    ;; Large-object manipulations _must_ occur within a transaction, since
    ;; the large object descriptors are only valid within the context of a
    ;; transaction.
    (syntax-rules ()
      ((with-pg-transaction con body ...)
       (let ((success #f))
         (dynamic-wind void
           (lambda ()
             (begin0
               (begin (pg-exec con "BEGIN WORK") body ...)
               (set! success #t)))
           (lambda ()
             (pg-exec con (if success "COMMIT WORK" "ROLLBACK WORK"))))))))

  (define (pg-for-each conn select-form callback)
    ;; Create a cursor for SELECT-FORM, and call CALLBACK for each result.
    ;; Uses the PostgreSQL database connection CONN. SELECT-FORM must be an
    ;; SQL SELECT statement. The cursor is created using an SQL DECLARE
    ;; CURSOR command, then results are fetched successively until no results
    ;; are left. The cursor is then closed.
    ;;
    ;; The work is performed within a transaction. The work can be
    ;; interrupted before all tuples have been handled by RAISEing a tag
    ;; called 'pg-finished.
    (let ((cursor (symbol->string (gensym "pgelcursor"))))
      (with-handlers (((lambda (tag) (eq? tag 'pg-finished)) void))  ;???
        (with-pg-transaction conn
          (pg-exec conn "DECLARE " cursor " CURSOR FOR " select-form)
          (dynamic-wind void 
           (lambda ()
             (let loop ()
               (let ((res
                      (pg-result (pg-exec conn "FETCH 1 FROM " cursor)
                                 'tuples)))
                 (unless (zero? (length res))
                   (callback (first res))
                   (loop)))))
           (lambda () (pg-exec conn "CLOSE " cursor)))))))

  (define pg-connect
    (opt-lambda (dbname user (password "") (host "localhost") (port 5432))
      ;; Initiate a connection with the PostgreSQL backend.
      ;; Connect to the database DBNAME with the username USER,
      ;; on PORT of HOST, providing PASSWORD if necessary. Return a
      ;; connection to the database (as an opaque type).
      ;; If HOST is #f, attempt to connect to the database using a Unix socket.
      (let ((connection
             (PQ-setdb-login host (number->string port) #f #f dbname
                             user password)))
        (cond ((= (PQ-status connection) CONNECTION-BAD)
               (signal-authentication-failure (PQ-error-message connection))
               #f)
              (else
               (and (not *pg-disable-type-coercion*)  ;???
                    (null? *parsers*)
                    (initialize-parsers connection))
               (pg-exec connection "SET datestyle = 'ISO'")
               connection)))))

  (define (pg-exec connection . args)
    ;; Execute the SQL command given by the concatenation of ARGS
    ;; on the database to which we are connected via CONNECTION. Return
    ;; a result structure which can be decoded using `pg-result'.
    (let/ec return
      (let ((sql (apply string-append args)))
        (when (> (string-length sql) +MAX_MESSAGE_LEN+)  ;???
          (error "SQL statement too long" sql))
        (let ((result (PQ-exec connection sql)))
          (cond ((not result)
                 (signal-protocol-error "pg-exec command failed, no error code"))
                (else
                 (evcase (PQ-result-status result)
                   (PGRES-COMMAND-OK
                    (let ((status (PQ-result-status result)))  ;!!!
                      (PQ-clear result)
                      (make-pgresult connection status '()  ;!!!
                                     '())))  ;!!!
                   (PGRES-TUPLES-OK
                    (let* ((status (PQ-result-status result))  ;!!!
                           (row-count (PQ-ntuples result))
                           (attributes (read-attributes result))
                           (tuples
                            (build-list row-count
                                        (lambda (r)
                                          (read-tuple result r attributes)))))
                      (PQ-clear result)
                      (make-pgresult connection status attributes  ;!!!
                                     tuples)))  ;!!!
                   ;; ErrorResponse
                   (else
                    (let ((msg (PQ-result-error-message result)))
                      (signal-backend-error msg)
                      (PQ-clear result))))))))))  ;!!!

  (define (pg-result result what . args)
    ;; Extract WHAT component of RESULT.
    ;; RESULT should be a structure obtained from a call to `pg-exec',
    ;; and WHAT should be one of
    ;; 'connection -> return the connection object
    ;; 'status -> return the status string provided by the database
    ;; 'attributes -> return the metadata, as a list of lists
    ;; 'tuples -> return the data, as a list of lists
    ;; 'tuple n -> return the nth component of the data
    ;; 'oid -> return the OID (a unique identifier generated by PostgreSQL
    ;;         for each row resulting from an insertion"
    (cond ((eq? 'connection what) (pgresult-connection result))
          ((eq? 'status what) (pgresult-status result))
          ((eq? 'attributes what) (pgresult-attributes result))
          ((eq? 'tuples what) (pgresult-tuples result))
          ((eq? 'tuple what)
           (let ((which
                  (if (pair? args) (car args) (error "which tuple?")))
                 (tuples (pgresult-tuples result)))
             (list-ref tuples which)))
          ((eq? 'oid what)
           (let ((status (pgresult-status result)))
             status  ;***
             #|(if (string-ci=? "INSERT" (substring status 0 6))  ;??? case-sensitive
                 (string->number (substring status 7 (string-index status #\space 7)))  ;!!!
                 (error "Only INSERT commands generate an oid: " status))|#))
          (else (error "Unknown result request: " what))))

  (define (pg-disconnect connection)
    (PQ-finish connection))

  ;; Attribute information provided by the backend is as follows:
  ;;   attribute-name (string)
  ;;   attribute-type as an oid from table pg_type
  ;;   attribute-size (in bytes?)

  (define (read-attributes result)
    (let ((attribute-count (PQ-nfields result)))
      (build-list attribute-count
                  (lambda (c)
                    (let ((type-name (PQ-fname result c))
                          (type-id (PQ-ftype result c))
                          (type-len (PQ-fsize result c))
                          ;; This doesn't exist in the 6.3 protocol !!
                          (type-modifier (PQ-fmod result c)))
                      (list type-name type-id type-len))))))

  (define (read-tuple result rowno attributes)
    (let* ((num-attributes (length attributes))
           (tuples '()))
      (do ((colno 0 (+ colno 1))
           (type-ids (map cadr attributes) (cdr type-ids)))
          ((= colno num-attributes) (reverse! tuples))
        (cond ((PQ-getisnull result rowno colno)
               (set! tuples (cons '() tuples)))  ;??? #f, 'NULL
              (else
               (let* ((raw (PQ-getvalue result rowno colno))
                      (parsed (parse raw (car type-ids))))
                 (set! tuples (cons parsed tuples))))))))

  ;; Type coercion support ==============================================
  ;;
  ;; When returning data from a SELECT statement, PostgreSQL starts by
  ;; sending some metadata describing the attributes. This information
  ;; is read by `READ-ATTRIBUTES', and consists of each attribute's
  ;; name (as a string), its size (in bytes), and its type (as an oid
  ;; which points to a row in the PostgreSQL system table pg_type). Each
  ;; row in pg_type includes the type's name (as a string).
  ;;
  ;; We are able to parse a certain number of the PostgreSQL types (for
  ;; example, numeric data is converted to a numeric Scheme type,
  ;; dates are converted to DrScheme date representation, booleans to
  ;; Scheme booleans). However, there isn't a fixed mapping from a
  ;; type to its OID which is guaranteed to be stable across database
  ;; installations, so we need to build a table mapping OIDs to parser
  ;; functions.
  ;;
  ;; This is done by the procedure `INITIALIZE-PARSERS', which is run
  ;; the first time a connection is initiated with the database from
  ;; this invocation of DrScheme, and which issues a SELECT statement to
  ;; extract the required information from pg_type. This initialization
  ;; imposes a slight overhead on the first request, which you can avoid
  ;; by setting `*PG-DISABLE-TYPE-COERCION*' to #t if it bothers you.
  ;; ====================================================================

  ;; See `man pgbuiltin' for details on PostgreSQL builtin types

  (define (integer-parser str) (string->number str))  ;!!!

  (define (float-parser str) (string->number str))  ;!!!

  ;; FIXME this may need support for charset decoding
  (define (text-parser str) str)

  (define (bool-parser str)
    (cond ((string=? "t" str) #t)
          ((string=? "f" str) #f)
          (else (error "Badly formed boolean from backend: " str))))

  (define (parse-timestamp str)
    (let* ((year (string->number (substring str 0 4)))
           (month (string->number (substring str 5 7)))
           (day (string->number (substring str 8 10)))
           (hours (string->number (substring str 11 13)))
           (minutes (string->number (substring str 14 16)))
           (seconds (string->number (substring str 17 19)))
           (start-tz (if (eqv? #\+ (string-ref str (- (string-length str) 3)))
                         (- (string-length str) 3)
                         #f))
           (tz (when start-tz
                 (string->number (substring str start-tz))))
           (milliseconds
            (if (eqv? (string-ref str 19) #\.)
                (string->number (substring str 20
                                           (or start-tz (string-length str))))
                0)))
      (values year month day hours minutes seconds milliseconds tz)))

  ;; Format for abstime/timestamp etc with ISO output syntax is
  ;;
  ;;    "1999-01-02 05:11:23.0345645+01"
  ;;
  ;; which we convert to a MzScheme date record. Thanks to James Anderson
  ;; for a fix for timestamp format in PostgreSQL 7.3 (with or without
  ;; tz, with or without milliseconds).

  (define (timestamp-parser str)
    (let-values (((year month day hours minutes seconds msec tz)
                  (parse-timestamp str)))
      (seconds->date (find-seconds seconds minutes hours day month year))))

  (define (precise-timestamp-parser str)
    (let-values (((year month day hours minutes seconds milliseconds tz)
                  (parse-timestamp str)))
      (seconds->date (+ (find-seconds seconds minutes hours day month year)
                        (/ milliseconds 1000.0)))))

  ;; An interval is what you get when you subtract two timestamps. We
  ;; convert to a number of seconds.

  (define (interval-parser str)
    (let* ((hours (string->number (substring str 0 2)))
           (minutes (string->number (substring str 3 5)))
           (seconds (string->number (substring str 6 8)))
           (milliseconds (string->number (substring str 9))))
      (+ (/ milliseconds (expt 10.0 (- (string-length str) 9)))
         seconds
         (* 60 minutes)
         (* 60 60 hours))))

  ;; Format for abstime/timestamp etc with ISO output syntax is
  ;;   "1999-01-02 00:00:00+01"
  ;; which we convert to a MzScheme date record

  (define (isodate-parser str)
    (let ((year (string->number (substring str 0 4)))
          (month (string->number (substring str 5 7)))
          (day (string->number (substring str 8 10)))
          (hours (string->number (substring str 11 13)))
          (minutes (string->number (substring str 14 16)))
          (seconds (string->number (substring str 17 19)))
          (tz (string->number (substring str 19 22))))  ;!!!
      (seconds->date (find-seconds seconds minutes hours day month year))))

  ;; Format for date with ISO output syntax is
  ;;;    "1999-01-02"
  ;; which we convert to a MzScheme date record.

  (define (date-parser str)
    (let ((year (string->number (substring str 0 4)))
	  (month (string->number (substring str 5 7)))
	  (day (string->number (substring str 8 10))))
      (seconds->date (find-seconds 0 0 0 day month year)))) 
 
  (define type-parsers
    `(("bool"      . ,bool-parser)
      ("char"      . ,text-parser)
      ("char2"     . ,text-parser)
      ("char4"     . ,text-parser)
      ("char8"     . ,text-parser)
      ("char16"    . ,text-parser)
      ("text"      . ,text-parser)
      ("varchar"   . ,text-parser)
      ("numeric"   . ,integer-parser)
      ("int2"      . ,integer-parser)
      ("int4"      . ,integer-parser)
      ("int8"      . ,integer-parser)
      ("oid"       . ,integer-parser)
      ("float4"    . ,float-parser)
      ("float8"    . ,float-parser)
      ("money"     . ,text-parser)       ;"$12.34"
      ("abstime"   . ,timestamp-parser)
      ("date"      . ,date-parser)
      ("timestamp" . ,timestamp-parser)
      ("timestamptz" . ,timestamp-parser)
      ("datetime"  . ,timestamp-parser)
      ("time"      . ,text-parser)     ;preparsed "15:32:45"
      ("timetz"    . ,text-parser)
      ("reltime"   . ,text-parser)     ;don't know how to parse these
      ("timespan"  . ,text-parser)
      ("interval"  . ,interval-parser)
      ("tinterval" . ,text-parser)))

  (define (initialize-parsers connection)
    (let* ((pgtypes (pg-exec connection "SELECT typname,oid FROM pg_type"))
           (tuples (pg-result pgtypes 'tuples)))
      (set! *parsers* '())
      (for-each  ;!!! map
       (lambda (tuple)
         (let* ((typname (first tuple))
                (oid (string->number (second tuple)))
                (type (assoc typname type-parsers)))
           (when (pair? type)
             (set! *parsers* (cons (cons oid (cdr type)) *parsers*)))))
       tuples)))

  ;; STR is an attribute's value retrieved as a string from the backend.
  ;; OID is the associated oid returned with the metadata. If we find a
  ;; parser for this type, we apply it, otherwise just return the
  ;; unparsed value.

  (define (parse str oid)
    (let ((parser (assv oid *parsers*)))
      (if (pair? parser)
          ((cdr parser) str)
          str)))

  (define (pg-notices conn)
    (let ((notify (PQ-notifies conn)))
      (if (eq? notify #f)
          '()
          (let ((notice (PG-notify-relname notify)))
            (PQ-free-notify notify)
            (cons notice (pg-notices conn))))))
  
  ;;; Large object support ================================================
  ;;;
  ;;; Sir Humphrey: Who is Large and to what does he object?
  ;;;
  ;;; Large objects are the PostgreSQL way of doing what most databases
  ;;; call BLOBs (binary large objects). In addition to being able to
  ;;; stream data to and from large objects, PostgreSQL's
  ;;; object-relational capabilities allow the user to provide functions
  ;;; which act on the objects.
  ;;;
  ;;; For example, the user can define a new type called "circle", and
  ;;; define a C function called `circumference' which will act on
  ;;; circles. There is also an inheritance mechanism in PostgreSQL. 
  ;;;
  ;;; The PostgreSQL large object interface is similar to the Unix file
  ;;; system, with open, read, write, lseek etc.
  ;;;======================================================================

  ;; Returns an OID

  (define pglo-create
    (opt-lambda (connection (modestr "r"))
      (let* ((mode (cond ((integer? modestr) modestr)
                         ((string-ci=? "r" modestr) INV-READ)
                         ((string-ci=? "w" modestr) INV-WRITE)
                         ((string-ci=? "rw" modestr)
                          (bitwise-ior INV-READ INV-WRITE))
                         (else (error "Bad mode" modestr))))
             (oid (lo-creat connection mode)))
        (cond ((not (integer? oid))
               (signal-backend-error "Didn't return an OID"))
              ((zero? oid)
               (signal-backend-error "Can't create large object"))
              (else oid)))))

  ;; Modestring: default "r", or "w" or "rw".
  ;; Returns a file descriptor for use in later lo-* procedures        

  (define pglo-open
    (opt-lambda (connection oid (modestr "r"))
      (let* ((mode (cond ((integer? modestr) modestr)
                         ((string-ci=? "r" modestr) INV-READ)
                         ((string-ci=? "w" modestr) INV-WRITE)
                         ((string-ci=? "rw" modestr)
                          (bitwise-ior INV-READ INV-WRITE))
                         (else (error "Bad mode " modestr))))
             (fd (lo-open connection oid mode)))
        (unless (integer? fd)
          (error 'pglo-open "Result not integer: " fd))
        fd)))

  (define (pglo-close connection fd)
    (lo-close connection fd))

  (define (pglo-read connection fd bytes)
    (let ((buf (make-string bytes)))
      (lo-read connection fd buf bytes)
      buf))

  (define (pglo-write connection fd buf)
    (lo-write connection fd buf (string-length buf)))

  (define (pglo-lseek connection fd offset whence)
    (lo-lseek connection fd offset whence))

  (define (pglo-tell connection oid)
    (lo-tell connection oid))

  (define (pglo-unlink connection oid)
    (lo-unlink connection oid))

  (define (pglo-import connection filename)
    (lo-import connection filename))

  (define (pglo-export connection oid filename)
    (lo-export connection oid filename))

  ;; DBMS metainformation ================================================
  ;;
  ;; Metainformation such as the list of databases present in the
  ;; database management system, list of tables, attributes per table.
  ;; This information is not available directly, but can be deduced by
  ;; querying the system tables.
  ;;
  ;; Based on the queries issued by psql in response to user commands
  ;; `\d' and `\d tablename'; see file pgsql/src/bin/psql/psql.c
  ;; ===================================================================

  (define (pg-databases connection)
    ;; Return a list of the databases available at this site.
    (let ((res (pg-exec connection "SELECT datname FROM pg_database")))
      (foldl append '() (pg-result res 'tuples))))

  (define (pg-tables connection)
    ;; Return a list of the tables present in this database.
    (let ((res
           (pg-exec connection "SELECT relname FROM pg_class, pg_user "
                    "WHERE relkind = 'r' AND relname !~ '^pg_' "
                    "AND usesysid = relowner ORDER BY relname")))
      (foldl append '() (pg-result res 'tuples))))

  (define (pg-columns connection table)
    ;; Return a list of the columns present in TABLE.
    (let ((res (pg-exec connection
                        (format "SELECT * FROM ~a WHERE 0 = 1" table))))
      (map car (pg-result res 'attributes))))

  (define (pg-backend-version conn)
    ;; Return a string identifying the version and operating environment
    ;; of the backend.
    (let ((res (pg-exec conn "SELECT version()")))
      (first (pg-result res 'tuple 0))))

  ;; support routines ===================================================

  ;; Double any ' characters. What else needs quoting?

  (define (pg-quote obj)  ;***
    (cond ((number? obj) (number->string obj))
          ((char? obj) (string #\' obj #\'))
          ((eq? #t obj) "'t'")
          ((eq? #f obj) "'f'")
          ((string? obj) (string-append "'" (string-replace "'" "''" obj) "'"))
          (else (error "Don't know how to quote this" obj))))

  ;; replace all occurrences of OLD by NEW in the string STR
  (define (string-replace old new str)
    ;; I'm lazy
    (regexp-replace* old str new)))
  
;;; *EOF*
