;;; -*- Scheme -*-

;;;; Demonstration & testing of DrScheme pg module

;;; Copyright (C) 2001  Zbigniew Jurkiewicz

;;; This file runs a limited number of tests on the database code.  It
;;; assumes you have a database user "scott" either with no password
;;; or with password "tiger"; you can change the username and
;;; password to something more appropriate.  It uses a special "test"
;;; database (which you will have to create first, either using
;;; `create_database' from the commandline, or from psql using `CREATE
;;; DATABASE'), and should clean up after itself.

;;; These tests assume that a table named "test" is defined in the
;;; system catalog, and that the user identified in
;;; CALL-WITH-TEST-CONNECTION has the rights to access that table. 

(current-library-collection-paths
 '("~zbyszek/plt/collects" "~zbyszek/local-collects"))

(require (lib "list.ss")) 
(require (lib "date.ss"))
(require (lib "process.ss"))
(require (lib "assert.scm" "zbyszek"))
(require (lib "mzpg.scm" "libpq" "mzpg"))

;;; !!! CHANGE THE VALUES HERE !!!

(define (call-with-test-connection function)
  (with-pg-connection (conn "test" "zbyszek" "" "localhost")
    (function conn)))

(define-syntax with-test-connection
  (syntax-rules ()
    ((with-test-connection (conn) body ...)
     (call-with-test-connection (lambda (conn) body ...)))))
 
(define (test-simple)
  (let ((*pg-disable-type-coercion* #t))  ;!!!
    (with-test-connection (conn)
      (printf "backend ~a~%" (pg-backend-version conn)))))

(define (test-general)
  (with-test-connection (conn)
    (printf "Running pg.scm tests against backend ~a~%"
            (pg-backend-version conn))
    ;; client encoding supported since PostgreSQL v7.1
    (printf "Client encoding is ~A~%" (pg-client-encoding conn))
    (printf "Date style is ~A~%" (pg-date-style conn))
    (let* ((r2 (pg-exec conn "CREATE TABLE pgscmtest (a int, b float, c money)"))
           (r3 (pg-exec conn
                        "INSERT INTO pgscmtest VALUES (3, -1234.5e67, '$123.45')"))
           (r4 (pg-exec conn "SELECT * FROM pgscmtest"))  ;***
           (r5 (pg-exec conn "DROP TABLE pgscmtest")))
      (printf "~%==============================================~%")
      (printf "status of CREATE is ~s~%" (pg-result r2 'status))
      (printf "status of INSERT is ~s~%" (pg-result r3 'status))
      (printf "oid of INSERT is ~s~%" (pg-result r3 'oid))
      (printf "status of SELECT is ~s~%" (pg-result r4 'status))  ;***
      (printf "attributes of SELECT are ~s~%" (pg-result r4 'attributes))  ;***
      (printf "tuples of SELECT are ~s~%" (pg-result r4 'tuples))  ;***
      (printf "first tuple of SELECT is ~s~%" (pg-result r4 'tuple 0))  ;***
      (printf "status of DROP is ~s~%" (pg-result r5 'status))
      (printf "==============================================~%"))))
 
(define (test-insert)
  (with-test-connection (conn)
    (let ((res #f)
          (count 0)
          (created #t))
      (dynamic-wind void
        (lambda ()
          (pg-exec conn "CREATE TABLE count_test(key int, val int)")
          (do ((i 1 (+ i 1))) ((> i 100))
            (pg-exec conn
                     (format "INSERT INTO count_test VALUES(~s, ~s)"
                             i (* i i))))
          (set! res (pg-exec conn "SELECT count(val) FROM count_test"))
          (unless (= 100 (first (pg-result res 'tuple 0)))  ;!!! assert i bez wypisywania
            (printf "Insert test failed!~%"))
          (set! res (pg-exec conn "SELECT sum(key) FROM count_test"))
          (unless (= 5050 (first (pg-result res 'tuple 0)))  ;!!! jw.
            (printf "Insert test failed!~%"))
          ;; This iterator does the equivalent of the sum(key) SQL statement
          ;; above, but on the client side. 
          (pg-for-each conn "SELECT key FROM count_test"
                       (lambda (tuple) (set! count (+ count (first tuple)))))  ;!!!
          (unless (= 5050 count)  ;!!! jw.
            (printf "Iterator test failed!~%")))
        (lambda ()
          (when created
            (pg-exec conn "DROP TABLE count_test")))))))

(define (test-date)
  (with-test-connection (conn)
    (let ((created #f))
      (dynamic-wind void
        (lambda ()
          (pg-exec conn
                   "CREATE TABLE pgscmtest (a timestamp, b abstime, c time, d date)")
          (pg-exec conn "INSERT INTO pgscmtest VALUES "
                   "(current_timestamp, 'now', 'now', 'now')")
          (set! created #t)
          (let* ((res (pg-exec conn "SELECT * FROM pgscmtest"))
                 (parsed (first (pg-result res 'tuples))))
            (printf "attributes ~a~%" (pg-result res 'attributes))
            (printf "Timestamp = ~s~%abstime = ~s~%time = ~s~%date = ~s~%"
                    (date->string (first parsed) #t)
                    (date->string (second parsed) #t)
                    (third parsed)
                    (date->string (fourth parsed)))))
        (lambda ()
          (when created
            (pg-exec conn "DROP TABLE pgscmtest")))))))

(define (test-metadata)
  (with-test-connection (conn)
    (pg-exec conn "CREATE TABLE pgscmmetatest (a int, b VARCHAR(4))")  ;***
    (let ((dbs (pg-databases conn))
          (tables (pg-tables conn)))
      (printf "List of databases: ~s~%" dbs)
      (printf "List of tables in test database: ~s~%" tables)
      (printf "List of columns in the table ~s: ~s~%"
              (car tables) (pg-columns conn (car tables))))
    (pg-exec conn "DROP TABLE pgscmmetatest")))  ;***

(define (test-lo)
  (with-test-connection (conn)
    (with-pg-transaction conn
      (let* ((oid (pglo-create conn))
             (fd (pglo-open conn oid)))
        (sleep 1)
        (pglo-tell conn fd)
        (sleep 1)
        (pglo-unlink conn oid)))))

;; Test of large-object interface

(define (test-lo-read)
  (with-test-connection (conn)
    (with-pg-transaction conn
      (let* ((oid (pglo-create conn "rw"))
             (fd (pglo-open conn oid "rw")))
        (pglo-write conn fd "Hi there mate")
        (pglo-lseek conn fd 3 0)           ;SEEK_SET = 0
        (unless (= 3 (pglo-tell conn fd))
          (error "Wrong position in TEST-LO-READ"))
        ;; This should print "there mate"
        (fprintf (current-error-port) "Read ~s from lo~%"
                 (pglo-read conn fd 10))
        (pglo-close conn fd)
        (pglo-unlink conn oid)))))
 
(define (test-lo-import)
  (with-test-connection (conn)
    (with-pg-transaction conn
      (let ((oid (pglo-import conn "/etc/group")))
        (pglo-export conn oid "/tmp/group")
        (cond ((system "diff /tmp/group /etc/group")
               (fprintf (current-error-port) "pglo-import test succeeded~%")
               (delete-file "/tmp/group"))
              (else
               (fprintf (current-error-port)
                        "pglo-import test failed: check differences
 between files /etc/group and /tmp/group")))
        (pglo-unlink conn oid)))))

(define (test-notifications)
  (with-test-connection (conn)
    (let ((res (pg-exec conn "LISTEN pg_test_listen")))
      (printf "LISTEN -> ~S~%" (pg-result res 'status))
      (unless (null? (pg-notices conn))  ;!!!
        (fprintf (current-error-port) "Error in LISTEN~%"))
      (pg-exec conn "SELECT * FROM pg_type")
      (assert (null? (pg-notices conn)))
      (set! res (pg-exec conn "NOTIFY pg_test_listen"))
      (printf "NOTIFY -> ~S~%" (pg-result res 'status))
      (printf "In TEST-NOTIFICATIONS notices are ~S~%"
              (pg-notices conn)))))

(define (test)
   (test-simple)
   (test-general)
   (test-insert)
   (test-date)
   (test-metadata)
   (test-lo)
   (test-lo-read)
   (test-lo-import)
   (test-notifications))

;;; *EOF*
