;;; Project:           Scheme-PG
;;; Author:            David J. Neu, djneu@acm.org
;;; Maintainer:        David J. Neu, djneu@acm.org
;;; Project Home Page: http://scheme-pg.sourceforge.net
;;; Copyright:         Copyright (c) 2004 Universal Technical Resource Services, Inc.
;;; License:           MIT License, see license.txt
;;; CVS Id:            $Id: examples.ss,v 1.3 2004/09/16 20:50:09 djneu Exp $

(require (lib "errortrace.ss" "errortrace"))
(require (lib "scheme-pg.scm" "libpq" "scheme-pg"))

(printf "Before running this program please comment out the call to (exit) and~n")
(printf "notice that the program attempts to drop a table pers, and it creates~n")
(printf "a table pers in a user specified db.  This safeguard is included to~n")
(printf "avoid unwanted changes to your DBMS.~n")

;;(exit)

(define (prompt aprompt)
  (printf "~a: " aprompt)
  (read))

(define ldbname (prompt "dbname"))
(define lusername (prompt "username"))
(define lpassword (prompt "password"))

(define lconnection
  (connection-open (format "dbname=~a user=~a password=~a"
                           ldbname lusername lpassword)))

;;; Example 1: Set up a simple database table and insert some records
;;; using all four insert syntax

;;; Drop table if it exists, but don't bomb if it doesn't.

(with-handlers
    ((exn?
      (lambda (e) (printf "table pers does not exist - continuing ...~n"))))
  (execute-sql lconnection "DROP table pers"))

(execute-sql lconnection
"CREATE TABLE pers
(id serial PRIMARY KEY,
 \"first-name\" character varying(25),
 \"last-name\" varchar(25),
 age float4,
 birthday timestamp without time zone);")

(execute-sql lconnection
             (insert pers (1 "Mary" "Smith" 20 "1983-10-12 02:50:41")))

(execute-sql lconnection
             (insert pers ((id 2) (first-name "John") (last-name "Doe") (age 40)
                           (birthday "1960-01-01 12:30:04"))))

(execute-sql lconnection
             (insert pers ((id . 3) (first-name . "Mary") (last-name . "Doe")
                           (age . 36) (birthday . "1965-02-01 17:55:44"))))

(execute-sql lconnection (insert pers (id first-name last-name age birthday)
                                 (4 "Baby" "Doe" 1.5, "2003-03-01 01:02:03")))

;;; Example 2: Use a result to display the rows retrieved by a query

(define lresult (result-open lconnection (select all (pers))))

(let loop ((llength (result-length lresult))
           (i 0))
  (when (> llength 0)
    (printf "~a~n" (result-ref lresult i))
    (loop (- llength 1) (+ i 1))))

#|
(result-close lresult)

(newline) (newline)

;;; Example 3: Use a cursor to display the rows retrieved by a query

(define lcursor
  (cursor-open lconnection
               (order-by (select all (pers))
                         ((last-name asc) (first-name desc)))
               'server))

(let loop ((lrow (cursor-next lcursor)))
  (unless (eoc-object? lrow)
    (printf "~a ~a ~a ~a ~a~n"
            (cdr (assq 'id lrow))
            (cdr (assq 'first-name lrow))
            (cdr (assq 'last-name lrow))
            (cdr (assq 'age lrow))
            (cdr (assq 'birthday lrow)))
    (loop (cursor-next lcursor))))

(cursor-close lcursor)

(newline) (newline)

;;; Example 4: Use a cursor to create a list of the rows retrieved by
;;; a query

(define query-four
  (order-by
   (select all (pers)
           (where (like first-name "M%")))
   ((age asc))))

(define (loop-forward aconnection)
  (let ((lcursor (cursor-open aconnection query-four 'client)))
    (let loop ((lrow (cursor-first lcursor)))
      (if (eoc-object? lrow)
          (begin
            (cursor-close lcursor)
            '())
          (cons lrow (loop (cursor-next lcursor)))))))

(display (loop-forward lconnection))
(newline) (newline)

;;; Example 5: Same as the previous example except it loops through
;;; the rows in reverse order

(define (loop-backward aconnection)
  (let ((lcursor (cursor-open aconnection query-four 'client)))
    (let loop ((lrow (cursor-last lcursor)))
      (if (eoc-object? lrow)
          (begin
            (cursor-close lcursor)
            '())
          (cons lrow (loop (cursor-previous lcursor)))))))

(display (loop-backward lconnection))
(newline) (newline)

;;; Example 6: Use the list-like syntax of a stream to display the
;;; rows retrieved by a query

(define lstream
  (stream lconnection (order-by (select all (pers)) ((id asc))) 'server))

(let loop ((lstream lstream))
  (unless (stream-null? lstream)
    (printf "~a~n" (stream-car lstream))
    (loop (stream-cdr lstream))))

(newline) (newline)

;;; Example 7: Use execute-sql to INSERT, UPDATE, DELETE, show the
;;; number of records affected and demonstrate the advantages of
;;; quasiquoting.

(let ((lnumber-rows
       (execute-sql lconnection
                    (insert pers (5, "Joe" "Smith" 25 "1982-12-15 14:20:01")))))
  (printf "~a rows inserted~n" lnumber-rows))

(let ((lnumber-rows
       (execute-sql lconnection
                    (update pers ((first-name "Jospeh"))
                            (where (= (pers id) 5))))))
  (printf "~a rows updated~n" lnumber-rows))

(define (get-first-name)
  "Mary")

(let ((lnumber-rows
       (execute-sql lconnection
                    (delete pers (where (and (= first-name ,(get-first-name))
                                             (like last-name "Smi%")))))))
  (printf "~a rows deleted~n" lnumber-rows))

(newline) (newline)

;;; Example 8: Demonstrate distinct and limit.

(define lstream
  (stream lconnection (distinct (select (last-name) (pers))) 'server))

(let loop ((lstream lstream))
  (unless (stream-null? lstream)
    (printf "~a~n" (stream-car lstream))
    (loop (stream-cdr lstream))))

(newline) (newline)

(define lstream
  (stream lconnection
          (limit (order-by (select all (pers)) ((id asc))) (2 1))
          'server))

(let loop ((lstream lstream))
  (unless (stream-null? lstream)
    (printf "~a~n" (stream-car lstream))
    (loop (stream-cdr lstream))))
|#

(connection-close lconnection)

;;; *EOF*
