#|
// Project:           Scheme-PG
// Author:            David J. Neu, djneu@acm.org
// Project Home Page: http://scheme-pg.sourceforge.net
// Copyright:         Copyright (c) 2004 Universal Technical Resource Services, Inc.
// License:           MIT License, see license.txt
//
// Modified by Zbigniew Jurkiewicz to use libpq collection instead of FFI.
|#

;;; !!! Zamaskowane bo brak PQftable w Postgresie

#|
See s-pg-types.h for data type constants. The file s-pg-types.h was
manually constructed from the source mentioned.
|#

#|
Usage:
     
  # mzscheme
  > (require (lib "s-pg.ss" "scheme-pg"))
  > (define lconnection (connection-open "dbname=mydbname user=myuser password=mypassword"))
  > (define lresult (result-open lconnection "SELECT * FROM mytable"))
  > (display (result-ref lresult 0))
  > (result-close lresult)
|#

(module s-pg mzscheme
  (require (lib "class.ss"))
  (require (lib "etc.ss"))
  (require (lib "libpq-fe.scm" "libpq"))
  (require "s-pg-types.scm")
  (provide connection-open
           connection?
           connection-close
           execute-sql
           result-open
           result?
           ;;result-empty
           result-ref
           result-length
           result-metadata
           result-close
           escape-string)

  (define connection%
    (class object%
      (init-field cptr)
      (super-new)))

  (define (connection-open info)
    (unless (string? info)
      (raise-type-error 'connection-open "string" info))
    (let ((connection (PQ-connectdb info)))
      (when (= (PQ-status connection) CONNECTION-BAD)
        ;; exn:scheme-pg:connection-open
        (error (format "~aCONNECTION INFOMATION:  ~a"
                       (PQ-error-message connection)
                       info)))
      (new connection% (cptr connection))))

  (define (connection? object)
    (is-a? object connection%))

  (define (connection-close connection)
    (unless (connection? connection)
      (raise-type-error 'connection-close "connection" connection))
    (PQ-finish (get-field cptr connection))
    (void))

  (define (execute-sql connection query)
    (unless (connection? connection)
      (raise-type-error 'execute-sql "connection" connection))
    (unless (string? query)
      (raise-type-error 'execute-sql "string" query))
    (let ((result (PQ-exec (get-field cptr connection) query)))
      (unless (= (PQ-result-status result) PGRES-COMMAND-OK)
        (cond ((= (PQ-result-status result) PGRES-TUPLES-OK)
               ;; exn:scheme-pg:execute-sql:non-query
               (error (format "ERROR:  execute-sql should not be used to retrieve data.\nSQL STATEMENT:  ~a"
                              (PQ-result-error-message result)
                              query))  ;???
               (PQ-clear result))
              (else
               ;; exn:scheme-pg:execute-sql
               (error (format "~aSQL STATEMENT:  ~a"
                              (PQ-result-error-message result)
                              query))
               (PQ-clear result))))
      (let* ((info (PQ-cmd-tuples result))
             (number-rows-affected
              (if (equal? info "") 0 (string->number info))))
        (PQ-clear result)
        number-rows-affected)))

  (define result%
    (class object%
      (init-field cptr)
      (super-new)))
  
  ;; Creates an empty result

  #|
  (define result-empty
    (c-lambda (scheme-object) scheme-object
      "PGconn *lconnection;
       PGresult *lresult;

       if (!SCHEME_CPTRP(___arg1) ||
           strcmp(SCHEME_CPTR_TYPE(___arg1), \"connection\"))
         scheme_wrong_type(\"result-open\", \"c-ptr:connection\", -1, 1, ___arg1);
       else
         lconnection = (PGconn *)SCHEME_CPTR_VAL(aargv[0]);
       lresult = PQmakeEmptyPGresult(lconnection, NULL);
       return lresult;"))
  |#

  (define (result-open connection query)
    (unless (connection? connection)
      (raise-type-error 'result-open "connection" connection))
    (unless (string? query)
      (raise-type-error 'result-open "string" query))
    (let ((result (PQ-exec (get-field cptr connection) query)))
       (unless (= (PQ-result-status result) PGRES-TUPLES-OK)
         ;; exn:scheme-pg:result-open
         (error (format "~aSQL STATEMENT:  ~a"
                        (PQ-result-error-message result)
                        query))
         (PQ-clear result))
      (new result% (cptr result))))

  (define (result? object)
    (is-a? object result%))

  (define (result-ref result row-number)
    (unless (result? result)
      (raise-type-error 'result-ref "result" result))
    (unless (integer? row-number)
      (raise-type-error 'result-ref "integer" row-number))
    (let ((result (get-field cptr result)))
      ;; Note: this also raises an exception when the result is empty
      (when (or (< row-number 0) (> row-number (- (PQ-ntuples result) 1)))
        ;; exn:scheme-db:result-ref:index
        (error "index out of range"))
      (let loop ((row '()) (column-number (- (PQ-nfields result) 1)))
        ;; ROW is an association list consisting of pair's
        ;; Notice the descreasing order
        (if (>= column-number 0)
            (let* ((value
                    (if (PQ-getisnull result row-number column-number)
                        'null
                        (if (= 1 1)  ;;!!!tmp TO DO: DO else option if user wants result column values returned as strings
                            (dbms-to-scheme result row-number column-number 'GET-SCHEME-OBJECT)
                            (PQ-getvalue result row-number column-number))))
                   ;; PAIR is an length two improper list, with the car being
                   ;; the field name as a symbol and the cadr the field value
                   (pair (cons (string->symbol (PQ-fname result column-number))
                               value)))
              (loop (cons pair row) (- column-number 1)))
            row))))

  (define (sch-column-metadata result column-number type)
    ;; Metadata is an association list
    (let ((metadata
           (list
            ;; Need access to a connection so can run the following SELECT 
            ;; to retrieve the actual table name
            ;; "SELECT relname FROM pg_class WHERE oid=%ld\", PQ-ftable(result, column-number)
            ;;(cons 'table (string->number (PQ-ftable aresult column-number)))  ;!!!
            (cons 'column (string->symbol (PQ-fname result
                                                    column-number)))
            (cons 'type (string->symbol type)))))
      metadata))
  
  ;; The value of argument ACTION determines whether or not a Scheme object
  ;; or metadata is returned

  (define (dbms-to-scheme result row-number column-number action)
    (evcase (PQ-ftype result column-number)
      (BOOLOID
       (if (eq? action 'GET-SCHEME-OBJECT)
           (cond ((equal? (PQ-getvalue result row-number column-number) "t") #t)
                 ((equal? (PQ-getvalue result row-number column-number) "f") #f)
                 (else 
                  ;; exn:scheme-db:conversion:boolean
                  (error (format "Invalid boolean value in row ~a, column ~a of result"
                                 row-number column-number))))
	   (sch-column-metadata result column-number "boolean")))
      ;; Can't seem to represent the maximum bigint of 9223372036854775807
      ;; as a Scheme object
      (INT8OID
       (cond ((eq? action 'GET-SCHEME-OBJECT)
              (printf "bigint=~a\n" (PQ-getvalue result row-number column-number))
	      (string->number (PQ-getvalue result row-number column-number)))
             (else
              (sch-column-metadata result column-number "bigint"))))
      (BPCHAROID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "character")))
      (VARCHAROID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "character-varying")))
      (DATEOID
       (if (eq? action 'GET_SCHEME_OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "date")))
      (FLOAT8OID
       (if (eq? action 'GET-SCHEME-OBJECT)
           (string->number (PQ-getvalue result row-number column-number))
           (sch-column-metadata result column-number "double-precision")))
      ;; Also handles serial4
      (INT4OID
       (if (eq? action 'GET-SCHEME-OBJECT)
           (string->number (PQ-getvalue result row-number column-number))
           (sch-column-metadata result column-number "integer")))
      (FLOAT4OID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (string->number (PQ-getvalue result row-number column-number))
	   (sch-column-metadata result column-number "real")))
      (TEXTOID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "integer")))  ;???
      (TIMEOID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "time-without-time-zone")))
      (TIMETZOID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "time-with-time-zone")))
      (TIMESTAMPOID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "timestamp-without-time-zone")))
      (TIMESTAMPTZOID
       (if (eq? action 'GET-SCHEME-OBJECT)
	   (PQ-getvalue result row-number column-number)
           (sch-column-metadata result column-number "timestamp-with-time-zone")))
      (else
       ;; exn:scheme-db:conversion
       (error (format "Conversion of column ~a of result with OID=~a not handled"
                      column-number (PQ-ftype result column-number))))))
  
  (define (result-metadata result)
    (unless (result? result)
      (raise-type-error 'result-metadata "result" result))
    (let ((result (get-field cptr result))
          (metadata '()))
      (do ((i 0 (+ i 1)))
        ((>= i (PQ-nfields result)))
        (set! metadata
              (cons (dbms-to-scheme result -1 i 'GET-METADATA)
                    metadata)))
      metadata))

  (define (result-length result)
    (unless (result? result)
      (raise-type-error 'result-length "result" result))
    (PQ-ntuples (get-field cptr result)))

  (define (result-close result)
    (unless (result? result)
      (raise-type-error 'result-close "result%" result))
    (PQ-clear (get-field cptr result))
    (void))

  (define (escape-string object)
    (unless (string? object)
      (raise-type-error 'escape-string "string" string))
    (PQ-escape-string object)))

;;; *EOF*
