;;; testlibpq.scm
;;;
;;; Test the Scheme version of LIBPQ, the POSTGRES frontend library.

(require (lib "libpq-fe.scm" "libpq"))

(define (string-pad str n)
  (let ((len (string-length str)))
    (if (>= len 15)
        str
        (string-append str (make-string (- n len) #\space)))))

(define (testlibpq)
  ;; Begin by setting the parameters for a backend connection.  If the
  ;; parameters are false, then the system will try to use reasonable
  ;; defaults by looking up environment variables or, failing that,
  ;; using hardwired constants
  (let ((pghost #|"libdb"|# #f)  ;host name of the backend server
        (pgport #f)                ;port of the backend server
        (pgoptions #f)             ;special options to start up the backend
                                   ;server
        (pgtty #f)                 ;debugging tty for the backend server
        (db-name "template1"))
    ;; Make a connection to the database
    (let ((conn (PQ-setdb pghost pgport pgoptions pgtty db-name)))
      ;; Check to see that the backend connection was successfully made
      (cond ((= (PQ-status conn) CONNECTION-BAD)
             (fprintf (current-error-port)
                      "Connection to database '~a' failed.\n" db-name)
             (fprintf (current-error-port) "~a" (PQ-error-message conn)))
            (else (testlibpq-aux conn)))
      ;; Close the connection to the database and cleanup
      (PQ-finish conn))))

(define (testlibpq-aux conn)
  ;; Start a transaction block
  (let ((res (PQ-exec conn "BEGIN")))
    (cond ((not (= (PQ-result-status res) PGRES-COMMAND-OK))
           (fprintf (current-error-port) "BEGIN command failed\n")
           (PQ-clear res))
          (else
           ;; Should PQ-clear PGresult whenever it is no longer needed
           ;; to avoid memory leaks
           (PQ-clear res)
           ;; Fetch instances from the pg_database, the system catalog of
           ;; databases
           (set! res
                 (PQ-exec
                  conn
                  "DECLARE mycursor CURSOR FOR select * from pg_database"))
           (cond ((not (= (PQ-result-status res) PGRES-COMMAND-OK))
                  (fprintf (current-error-port)
                           "DECLARE CURSOR command failed\n")
                  (PQ-clear res))
                 (else
                  (PQ-clear res)
                  (set! res (PQ-exec conn "FETCH ALL IN mycursor"))
                  (cond ((not (= (PQ-result-status res) PGRES-TUPLES-OK))
                         (fprintf
                          (current-error-port)
                          "FETCH ALL command didn't return tuples properly\n")
                         (PQ-clear res))
                        (else
                         ;; First, print out the attribute names
                         (let ((n-fields (PQ-nfields res)))
                           (do ((i 0 (+ i 1)))
                             ((>= i n-fields))
                             (printf "~a" (string-pad (PQ-fname res i) 15)))
                           (printf "\n\n")
                           ;; Next, print out the instances
                           (do ((i 0 (+ i 1))) ((>= i (PQ-ntuples res)))
                             (do ((j 0 (+ j 1))) ((>= j n-fields))
                               (printf "~a"
                                       (string-pad (PQ-getvalue res i j) 15)))
                             (printf "\n")))
                         (PQ-clear res)
                         ;; Close the cursor
                         (set! res (PQ-exec conn "CLOSE mycursor"))
                         (PQ-clear res)
                         ;; Commit the transaction
                         (set! res (PQ-exec conn "COMMIT"))
                         (PQ-clear res)))))))))

;;; *EOF*
